# -*- coding: utf-8 -*-
"""
Copyright (c) 2011, Daniele Esposti <expo@expobrain.net>
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * The name of the contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""


class BitmapHandler(object):
    """
    Holds decode WebP image data and extra informations
    """
    RGB = 0
    RGBA = 1
    BGR = 2
    BGRA = 3
    YUV = 4

    FORMATS = (RGB, RGBA, BGR, BGRA, YUV)

    def __init__(self, bitmap, fmt, width, height, stride,
                 u_bitmap=None, v_bitmap=None, uv_stride= -1):
        """
        Constructor accepts the decode image data as a bitmap and its
        width/height.

        Passing a null image data, and invalid format or a non
        positive integer for width/height creates an instance to an invalid
        WebP image.

        If the image is in YUV format the bitmap parameter will be the Y(luma)
        component and the U/V chrominance component bitmap must be passed else
        the image will be invalid. The Y bitmap stride and the UV bitmap stride
        must be passed as well.

        :param bitmap: The image bitmap
        :param fmt: The image format
        :param width: The image width
        :param height: The mage height
        :param u_bitmap: The U chrominance component bitmap
        :param v_bitmap: The V chrominance component bitmap
        :param stride: The Y bitmap stride
        :param uv_stride: The UV stride

        :type bitmap: bytearray
        :type fmt: M{WebPImage.FORMATS}
        :type width: int
        :type height: int
        :type u_bitmap: bytearray
        :type v_bitmap: bytearray
        :type stride: int
        :type uv_stride: int
        """
        self.bitmap = bitmap
        self.u_bitmap = u_bitmap
        self.v_bitmap = v_bitmap
        self.stride = stride
        self.uv_stride = uv_stride
        self.format = fmt
        self.width = width
        self.height = height

        # Check if bitmap handler is valid
        is_valid = (isinstance(bitmap, bytearray)
                    and fmt in self.FORMATS
                    and width > -1
                    and height > -1)

        # Additional setups for YUV image
        if is_valid and fmt == self.YUV:
            # Check if YUV image is valid
            is_valid = (isinstance(u_bitmap, bytearray)
                        and isinstance(v_bitmap, bytearray)
                        and stride > -1
                        and uv_stride > -1)

        # Set valid flag
        self.is_valid = is_valid


class WebPHandlerError(IOError):
    pass


class WebPHandler(object):
    """
    Contains data relative to an WebP encoded image and allow loading and
    saving .webp files.

    The code is base on the documentation at
    http://code.google.com/speed/webp/docs/riff_container.html

    Public properties:

    * data        The WebP encoded image data
    * width       The image's width, -1 if the image s not valid
    * height      The image's height, -1 if the image s not valid
    * is_valid    True if the image's data is valid else False
    """

    @staticmethod
    def from_file(filename):
        """
        Load a .webp file and return the WebP handler

        :param filename: The file's name to be loaded
        :type filename: string
        :rtype: WebPHandler
        """
        from webm import decode

        data = file(filename, "rb").read()
        width, height = decode.GetInfo(data)

        return WebPHandler(bytearray(data), width, height)

    def __init__(self, data=None, width= -1, height= -1):
        """
        Constructor accepts the data, width and height of the WebP encoded
        image

        :param source: The image encoded data
        :param width: The image's width
        :param height: The image's height

        :type data: bytearray
        :type width: int
        :type height: int
        """
        # Public attributes
        self.data = data
        self.width = width
        self.height = height

    @property
    def is_valid(self):
        """
        Returns True if the current image is valid

        :rtype: bool
        """
        return self.data != None and (self.width > -1 and self.height > -1)



# -*- coding: utf-8 -*-
"""
Copyright (c) 2011, Daniele Esposti <expo@expobrain.net>
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * The name of the contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

from PIL import Image
from webm import yuv, decode
from webm.tests.common import WEBP_IMAGE_DATA, DECODE_FILENAME

try:
    import unittest2 as unittest
except ImportError:
    import unittest


class YuvTests(unittest.TestCase):
    """
    YUV to RGB conversion module test case
    """

    def test_init(self):
        """
        Test intialise lookup tables
        """
        import webm.yuv as yuv

        self.assertEqual(len(yuv.VP8kVToR), 256)
        self.assertEqual(len(yuv.VP8kUToB), 256)
        self.assertEqual(len(yuv.VP8kVToG), 256)
        self.assertEqual(len(yuv.VP8kUToG), 256)
        self.assertEqual(
            len(yuv.VP8kClip), yuv.YUV_RANGE_MAX - yuv.YUV_RANGE_MIN)

        self.assertNotEqual(yuv.VP8kVToR, [0] * 256)
        self.assertNotEqual(yuv.VP8kUToB, [0] * 256)
        self.assertNotEqual(yuv.VP8kVToG, [0] * 256)
        self.assertNotEqual(yuv.VP8kUToG, [0] * 256)
        self.assertNotEqual(
            yuv.VP8kClip, [0] * (yuv.YUV_RANGE_MAX - yuv.YUV_RANGE_MIN))

    def test_output_YUV_to_RGB(self):
        """
        Export DecodeYUV() method result to a RGB file
        """
        # Get YUV data and convert to RGB
        result = decode.DecodeYUV(WEBP_IMAGE_DATA)
        result = yuv.YUVtoRGB(result)

        # Save image
        image = Image.frombuffer(
            "RGB", (result.width, result.height), str(result.bitmap),
            "raw", "RGB", 0, 1
        )
        image.save(DECODE_FILENAME.format("YUV_RGB"))

    def test_output_YUV_to_RGBA(self):
        """
        Export DecodeYUV() method result to a RGBA file
        """
        # Get YUV data and convert to RGB
        result = decode.DecodeYUV(WEBP_IMAGE_DATA)
        result = yuv.YUVtoRGBA(result)

        # Save image
        image = Image.frombuffer(
            "RGBA", (result.width, result.height), str(result.bitmap),
            "raw", "RGBA", 0, 1
        )
        image.save(DECODE_FILENAME.format("YUV_RGBA"))

    def test_output_YUV_to_BGR(self):
        """
        Export DecodeYUV() method result to a BGR file
        """
        # Get YUV data and convert to BGR
        result = decode.DecodeYUV(WEBP_IMAGE_DATA)
        result = yuv.YUVtoBGR(result)

        # Save image
        image = Image.frombuffer(
            "RGB", (result.width, result.height), str(result.bitmap),
            "raw", "BGR", 0, 1
        )
        image.save(DECODE_FILENAME.format("YUV_BGR"))

    def test_output_YUV_to_BGRA(self):
        """
        Export DecodeYUV() method result to a BGRA file
        """
        # Get YUV data and convert to BGRA
        result = decode.DecodeYUV(WEBP_IMAGE_DATA)
        result = yuv.YUVtoBGRA(result)

        # Save image
        image = Image.frombuffer(
            "RGBA", (result.width, result.height), str(result.bitmap),
            "raw", "BGRA", 0, 1
        )
        image.save(DECODE_FILENAME.format("YUV_BGRA"))



# -*- coding: utf-8 -*-
"""
Copyright (c) 2011, Daniele Esposti <expo@expobrain.net>
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * The name of the contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

from ctypes import (c_int, c_float, c_void_p, byref, memmove,
    create_string_buffer)
from webm import _LIBRARY
from webm.handlers import WebPHandler


# -----------------------------------------------------------------------------
# Exceptions
# -----------------------------------------------------------------------------

class EncodeError(Exception):
    """
    Exception for encoding errors
    """
    pass


# -----------------------------------------------------------------------------
# Internal functions
# -----------------------------------------------------------------------------

# Set argument types
_LIBRARY.WebPEncodeRGB.argtypes = [c_void_p, c_int, c_int, c_int,
                                   c_float, c_void_p]
_LIBRARY.WebPEncodeBGR.argtypes = [c_void_p, c_int, c_int, c_int, c_float,
                                   c_void_p]
_LIBRARY.WebPEncodeRGBA.argtypes = [c_void_p, c_int, c_int, c_int, c_float,
                                    c_void_p]
_LIBRARY.WebPEncodeBGRA.argtypes = [c_void_p, c_int, c_int, c_int, c_float,
                                    c_void_p]

# Set return types
_LIBRARY.WebPEncodeRGB.restype = c_int
_LIBRARY.WebPEncodeBGR.restype = c_int
_LIBRARY.WebPEncodeRGBA.restype = c_int
_LIBRARY.WebPEncodeBGRA.restype = c_int


def _encode(func, image, quality):
    """
    Encode the image with the given quality using the given encoding
    function

    :param func: The encoding function
    :param image: The image to be encoded
    :param quality: The encode quality factor

    :type function: function
    :type image: BitmapHandler
    :type quality: float
    """
    # Call encode function
    data = str(image.bitmap)
    width = c_int(image.width)
    height = c_int(image.height)
    stride = c_int(image.stride)
    q_factor = c_float(quality)
    output_p = c_void_p()

    size = func(data, width, height, stride, q_factor, byref(output_p))

    # Check return size
    if size == 0:
        raise EncodeError

    # Convert output
    output = create_string_buffer(size)

    memmove(output, output_p, size)

    return WebPHandler(bytearray(output), image.width, image.height)


# -----------------------------------------------------------------------------
# Public functions
# -----------------------------------------------------------------------------

def EncodeRGB(image, quality=100):
    """
    Encode the given RGB image with the given quality

    :param image: The RGB image
    :param quality: The encode quality factor

    :type image: BitmapHandler
    :type quality: float
    """
    return _encode(_LIBRARY.WebPEncodeRGB, image, quality)

def EncodeRGBA(image, quality=100):
    """
    Encode the given RGBA image with the given quality

    :param image: The RGBA image
    :param quality: The encode quality factor

    :type image: BitmapHandler
    :type quality: float
    """
    return _encode(_LIBRARY.WebPEncodeRGBA, image, quality)

def EncodeBGRA(image, quality=100):
    """
    Encode the given BGRA image with the given quality

    :param image: The BGRA image
    :param quality: The encode quality factor

    :type image: BitmapHandler
    :type quality: float
    """
    return _encode(_LIBRARY.WebPEncodeBGRA, image, quality)

def EncodeBGR(image, quality=100):
    """
    Encode the given BGR image with the given quality

    :param image: The BGR image
    :param quality: The encode quality factor

    :type image: BitmapHandler
    :type quality: float
    """
    return _encode(_LIBRARY.WebPEncodeBGR, image, quality)



# -*- coding: utf-8 -*-
"""
Copyright (c) 2011, Daniele Esposti <expo@expobrain.net>
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * The name of the contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

from handlers import BitmapHandler

"""
Porting of the YUVtoRGB converter code from the Chromium project. See
third_party/libwebp/yuv.c and /third_party/libwebp/yuv.h files for the original
source code.
"""


# -----------------------------------------------------------------------------
# Internal functions
# -----------------------------------------------------------------------------

# YUV conversion constants
YUV_FIX = 16                    # fixed-point precision
YUV_RANGE_MIN = -227            # min value of r/g/b output
YUV_RANGE_MAX = 256 + 226       # max value of r/g/b output
YUV_HALF = 1 << (YUV_FIX - 1)
VP8kVToR = [0] * 256
VP8kUToB = [0] * 256
VP8kVToG = [0] * 256
VP8kUToG = [0] * 256
VP8kClip = [0] * (YUV_RANGE_MAX - YUV_RANGE_MIN)


def _init_yuv_module():
    """
    Initialise the YUVtoRGB lookup tables
    """
    for i in xrange(256):
        VP8kVToR[i] = (89858 * (i - 128) + YUV_HALF) >> YUV_FIX
        VP8kUToG[i] = -22014 * (i - 128) + YUV_HALF
        VP8kVToG[i] = -45773 * (i - 128)
        VP8kUToB[i] = (113618 * (i - 128) + YUV_HALF) >> YUV_FIX

    for i in xrange(YUV_RANGE_MIN, YUV_RANGE_MAX):
        k = ((i - 16) * 76283 + YUV_HALF) >> YUV_FIX
        k = 0 if k < 0 else 255 if k > 255 else k

        VP8kClip[i - YUV_RANGE_MIN] = k


# Initialise YUVtoRGB lookup table on module loading
_init_yuv_module()


def _decode_YUV_image(image):
    """
    Decode the given image in YUV format to a RGB byte array

    :param image: The image in YUV format to be decoded
    :type image: BitmapHandler
    :rtype: bytearray
    """
    rgb_bitmap = bytearray()

    for h in xrange(image.height):
        for w in xrange(image.width):
            # Get luma
            i = h * image.stride + w
            y = image.bitmap[i]

            # WORKAROUND:
            # This is a workaround for YUV to RGB decoding to have a useful
            # grayscale image from the YUV data.
            # Remove this when the correct YUV to color RGB function is
            # found.
            rgb_bitmap.extend((y, y, y))

#            # Get chrominance
#            i = h * image.uv_stride + int(w/2)
#
#            if w % 2:
#                u = image.u_bitmap[i] & 0b1111
#                v = image.v_bitmap[i] & 0b1111
#            else:
#                u = image.u_bitmap[i] >> 4
#                v = image.v_bitmap[i] >> 4
#
#            # Calculate RGB values
#            r_off = VP8kVToR[v]
#            g_off = (VP8kVToG[v] + VP8kUToG[u]) >> YUV_FIX
#            b_off = VP8kUToB[u]
#
#            rgb_bitmap.append( VP8kClip[y + r_off - YUV_RANGE_MIN] )
#            rgb_bitmap.append( VP8kClip[y + g_off - YUV_RANGE_MIN] )
#            rgb_bitmap.append( VP8kClip[y + b_off - YUV_RANGE_MIN] )

    # End
    return rgb_bitmap


# -----------------------------------------------------------------------------
# Public functions
# -----------------------------------------------------------------------------

def YUVtoRGB(image):
    """
    Convert the given WebP image instance from a YUV format to an RGB format

    :param image: The WebP image in YUV format
    :type image: BitmapHandler
    :rtype: BitmapHandler
    """
    return BitmapHandler(
        _decode_YUV_image(image), BitmapHandler.RGB,
        image.width, image.height, image.width * 3
    )

def YUVtoRGBA(image):
    """
    Convert the given WebP image instance form YUV format to RGBA format

    :param image: The WebP image in YUV format
    :type image: BitmapHandler
    :rtype: BitmapHandler
    """
    rgb_bitmap = _decode_YUV_image(image)
    rgba_bitmap = bytearray()

    for i in xrange(len(rgb_bitmap) / 3):
        i *= 3

        rgba_bitmap.append(rgb_bitmap[i])
        rgba_bitmap.append(rgb_bitmap[i + 1])
        rgba_bitmap.append(rgb_bitmap[i + 2])
        rgba_bitmap.append(0xff)

    # Return the BitmapHandler in RGB format
    return BitmapHandler(
        rgba_bitmap, BitmapHandler.RGBA,
        image.width, image.height, image.width * 4
    )

def YUVtoBGR(image):
    """
    Convert the given WebP image instance form YUV format to BGR format

    :param image: The WebP image in YUV format
    :type image: BitmapHandler
    :rtype: BitmapHandler
    """
    rgb_bitmap = _decode_YUV_image(image)
    bgr_bitmap = bytearray()

    for i in xrange(len(rgb_bitmap) / 3):
        i *= 3

        bgr_bitmap.append(rgb_bitmap[i + 2])
        bgr_bitmap.append(rgb_bitmap[i + 1])
        bgr_bitmap.append(rgb_bitmap[i])

    # Return the BitmapHandler in BGR format
    return BitmapHandler(
        bgr_bitmap, BitmapHandler.BGR,
        image.width, image.height, image.width * 3
    )

def YUVtoBGRA(image):
    """
    Convert the given WebP image instance form YUV format to BGRA format

    :param image: The WebP image in YUV format
    :type image: BitmapHandler
    :rtype: BitmapHandler
    """
    rgb_bitmap = _decode_YUV_image(image)
    bgra_bitmap = bytearray()

    for i in xrange(len(rgb_bitmap) / 3):
        i *= 3

        bgra_bitmap.append(rgb_bitmap[i + 2])
        bgra_bitmap.append(rgb_bitmap[i + 1])
        bgra_bitmap.append(rgb_bitmap[i])
        bgra_bitmap.append(0xff)

    # Return the BitmapHandler in BGRA format
    return BitmapHandler(
        bgra_bitmap, BitmapHandler.BGRA,
        image.width, image.height, image.width * 4
    )



# -*- coding: utf-8 -*-
"""
Copyright (c) 2011, Daniele Esposti <expo@expobrain.net>
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * The name of the contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

from ctypes import (c_int, c_uint, create_string_buffer, memmove, c_void_p,
    POINTER)
from webm import _LIBRARY, PIXEL_ALPHA_SZ, PIXEL_SZ
from webm.handlers import BitmapHandler


# -----------------------------------------------------------------------------
# Exceptions
# -----------------------------------------------------------------------------
class HeaderError(Exception):
    """
    Exception for image header operations and manipulations
    """
    pass


# -----------------------------------------------------------------------------
# Internal functions
# -----------------------------------------------------------------------------

# Set function parameter types
_LIBRARY.WebPDecodeRGB.argtypes = [c_void_p, c_uint,
                                   POINTER(c_int), POINTER(c_int)]  # w, h
_LIBRARY.WebPDecodeRGBA.argtypes = [c_void_p, c_uint,
                                    POINTER(c_int), POINTER(c_int)]  # w, h
_LIBRARY.WebPDecodeBGR.argtypes = [c_void_p, c_uint,
                                   POINTER(c_int), POINTER(c_int)]  # w, h
_LIBRARY.WebPDecodeBGRA.argtypes = [c_void_p, c_uint,
                                   POINTER(c_int), POINTER(c_int)]  # w, h
_LIBRARY.WebPDecodeYUV.argtypes = [c_void_p, c_uint,
                                   POINTER(c_int), POINTER(c_int),  # w, h
                                   c_void_p, c_void_p,  # u, v
                                   POINTER(c_int),  # stride
                                   POINTER(c_int)]  # uv_stride
_LIBRARY.WebPGetInfo.argtypes = [c_void_p, c_uint,
                                 POINTER(c_int), POINTER(c_int)]  # w, h

# Set return types
_LIBRARY.WebPDecodeRGB.restype = c_void_p
_LIBRARY.WebPDecodeRGBA.restype = c_void_p
_LIBRARY.WebPDecodeBGR.restype = c_void_p
_LIBRARY.WebPDecodeBGRA.restype = c_void_p
_LIBRARY.WebPDecodeYUV.restype = c_void_p
_LIBRARY.WebPGetInfo.restype = c_uint


def _decode(data, decode_func, pixel_sz):
    """
    Decode the given WebP image data using given decode and with the given
    pixel size in bytes

    :param data: The original WebP image data
    :param decode_func: The decode function to be used
    :param pixel_sz: The pixel data size in bytes to calculate the decoded
                     image size buffer

    :type data: bytearray
    :type decode_func: function
    :type pixel_sz: int

    :rtype: tuple(bytearray, int, int)
    """
    # Prepare parameters
    width = c_int(-1)
    height = c_int(-1)
    size = len(data)

    # Decode image an return pointer to decoded data
    bitmap_p = decode_func(str(data), size, width, height)

    # Copy decoded data into a buffer
    width = width.value
    height = height.value
    size = width * height * pixel_sz
    bitmap = create_string_buffer(size)

    memmove(bitmap, bitmap_p, size)

    # End
    return (bytearray(bitmap), width, height)


# -----------------------------------------------------------------------------
# Public functions
# -----------------------------------------------------------------------------

def GetInfo(data):
    """
    Return the width and the height from the given WebP image data

    :param data: The original WebP image data
    :type data: bytearray
    :rtype: tuple(int, int)
    """
    # Call C function
    width = c_int(-1)
    height = c_int(-1)
    size = len(data)

    ret = _LIBRARY.WebPGetInfo(str(data), size, width, height)

    # Check return code
    if ret == 0:
        raise HeaderError

    # Return values
    return (width.value, height.value)


def DecodeRGB(data):
    """
    Decode the given WebP image data to a RGB bitmap

    :param data: The original WebP image data
    :type data: bytearray
    :rtype: WebPImage
    """
    bitmap, width, height = _decode(
        data, _LIBRARY.WebPDecodeRGB, PIXEL_SZ)

    return BitmapHandler(
        bitmap, BitmapHandler.RGB, width, height, PIXEL_SZ * width)


def DecodeBGR(data):
    """
    Decode the given WebP image data to a BGR bitmap

    :param data: The original WebP image data
    :type data: bytearray
    :rtype: WebPImage
    """
    bitmap, width, height = _decode(
        data, _LIBRARY.WebPDecodeBGR, PIXEL_SZ)

    return BitmapHandler(
        bitmap, BitmapHandler.BGR, width, height, PIXEL_SZ * width)

def DecodeBGRA(data):
    """
    Decode the given WebP image data to a BGRA bitmap

    :param data: The original WebP image data
    :type data: bytearray
    :rtype: WebPImage
    """
    bitmap, width, height = _decode(
        data, _LIBRARY.WebPDecodeBGRA, PIXEL_ALPHA_SZ)

    return BitmapHandler(
        bitmap, BitmapHandler.BGRA, width, height, PIXEL_ALPHA_SZ * width)

def DecodeRGBA(data):
    """
    Decode the given WebP image data to a RGBA bitmap

    :param data: The original WebP image data
    :type data: bytearray
    :rtype: WebPImage
    """
    bitmap, width, height = _decode(
        data, _LIBRARY.WebPDecodeRGBA, PIXEL_ALPHA_SZ)

    return BitmapHandler(
        bitmap, BitmapHandler.RGBA, width, height, PIXEL_ALPHA_SZ * width)

def DecodeYUV(data):
    """
    Decode the given WebP image data to a YUV bitmap

    :param data: The original WebP image data
    :type data: bytearray
    :rtype: WebPImage
    """
    # Prepare parameters
    width = c_int(-1)
    height = c_int(-1)
    size = len(data)
    u = create_string_buffer(0)
    v = create_string_buffer(0)
    stride = c_int(-1)
    uv_stride = c_int(-1)

    # Decode image an return pointer to decoded data
    bitmap_p = _LIBRARY.WebPDecodeYUV(
        str(data), size, width, height, u, v,
#        u, v,
        stride, uv_stride
    )

    # Convert data to Python types
    width = width.value
    height = height.value
    stride = stride.value
    uv_stride = uv_stride.value

    # Copy decoded data into a buffer
    size = stride * height
    bitmap = create_string_buffer(size)

    memmove(bitmap, bitmap_p, size)

    # Copy UV chrominace bitmap
    uv_size = uv_stride * height
    u_bitmap = create_string_buffer(uv_size)
    v_bitmap = create_string_buffer(uv_size)

    memmove(u_bitmap, u, uv_size)
    memmove(v_bitmap, v, uv_size)

    # End
    return BitmapHandler(
        bytearray(bitmap), BitmapHandler.YUV, width, height, stride,
        u_bitmap=bytearray(u_bitmap),
        v_bitmap=bytearray(v_bitmap),
        uv_stride=uv_stride
    )



# -*- coding: utf-8 -*-
"""
Copyright (c) 2011, Daniele Esposti <expo@expobrain.net>
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * The name of the contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

from webm.handlers import WebPHandler, BitmapHandler
from webm.tests.common import (IMAGE_WIDTH, IMAGE_HEIGHT, WEBP_IMAGE_DATA,
    WEBP_IMAGE_FILE)
import os

try:
    import unittest2 as unittest
except ImportError:
    import unittest


class BitmapHandlerTests(unittest.TestCase):
    """
    BitmapHandler tests cases
    """

    def test_image_types_enum(self):
        """
        Test image types enumerator
        """
        self.assertEqual(BitmapHandler.RGB, 0)
        self.assertEqual(BitmapHandler.RGBA, 1)
        self.assertEqual(BitmapHandler.BGR, 2)
        self.assertEqual(BitmapHandler.BGRA, 3)
        self.assertEqual(BitmapHandler.YUV, 4)

    def test_is_valid(self):
        """
        Test is_valid() property
        """
        # Invalid
        self.assertFalse(
            BitmapHandler(None, None, IMAGE_WIDTH, IMAGE_HEIGHT, -1).is_valid)
        self.assertFalse(BitmapHandler(None, None, -1, -1, -1).is_valid)
        self.assertFalse(
            BitmapHandler(None, BitmapHandler.YUV, 0, 0, None).is_valid)

        # Valid
        image = BitmapHandler(
            WEBP_IMAGE_DATA, BitmapHandler.RGB,
            IMAGE_WIDTH, IMAGE_HEIGHT, IMAGE_WIDTH * 3
        )

        self.assertTrue(image.is_valid)
        self.assertEqual(image.bitmap, WEBP_IMAGE_DATA)
        self.assertEqual(image.format, BitmapHandler.RGB)
        self.assertEqual(image.width, IMAGE_WIDTH)
        self.assertEqual(image.height, IMAGE_HEIGHT)

    def test_is_valid_yuv(self):
        """
        Test is_valid() property for YUV format
        """
        # Create fake Y and UV bitmaps
        bitmap = bytearray(IMAGE_WIDTH * IMAGE_HEIGHT)
        uv_bitmap = bytearray(int(IMAGE_WIDTH * IMAGE_HEIGHT / 2))

        # Create image instance
        image = BitmapHandler(
            bitmap, BitmapHandler.YUV, IMAGE_WIDTH, IMAGE_HEIGHT,
            u_bitmap=uv_bitmap, v_bitmap=uv_bitmap,
            stride=IMAGE_WIDTH, uv_stride=int(IMAGE_WIDTH / 2)
        )

        self.assertTrue(image.is_valid)
        self.assertEqual(image.bitmap, bitmap)
        self.assertEqual(image.format, BitmapHandler.YUV)
        self.assertEqual(image.width, IMAGE_WIDTH)
        self.assertEqual(image.height, IMAGE_HEIGHT)
        self.assertEqual(image.u_bitmap, uv_bitmap)
        self.assertEqual(image.v_bitmap, uv_bitmap)
        self.assertEqual(image.stride, IMAGE_WIDTH)
        self.assertEqual(image.uv_stride, int(IMAGE_WIDTH / 2))


class WebPHandlerTests(unittest.TestCase):
    """
    WebPHandler test cases
    """

    TEST_IMAGE_FILE = os.path.join(os.path.dirname(__file__),
                                   "webphandler_{0}.webp")

    def test_load_by_filename(self):
        """
        Test loading a .webp file
        """
        image = WebPHandler.from_file(WEBP_IMAGE_FILE)

        self.assertTrue(isinstance(image, WebPHandler))
        self.assertTrue(image.is_valid)
        self.assertEqual(image.width, IMAGE_WIDTH)
        self.assertEqual(image.height, IMAGE_HEIGHT)
        self.assertTrue(isinstance(image.data, bytearray))
